@@ -4,22 +4,68 @@ Support for multirange types adaptation.
 
 # Copyright (C) 2021 The Psycopg Team
 
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Any, Generic, overload
 from decimal import Decimal
-from typing import Any, Generic, List, Iterable
-from typing import MutableSequence, Optional, Type, Union, overload
 from datetime import date, datetime
+from functools import cache
+from collections.abc import Iterable, MutableSequence
 
+from .. import _oids
 from .. import errors as e
-from .. import postgres
+from .. import postgres, sql
 from ..pq import Format
-from ..abc import AdaptContext, Buffer, Dumper, DumperKey
-from ..adapt import RecursiveDumper, RecursiveLoader, PyFormat
+from ..abc import AdaptContext, Buffer, Dumper, DumperKey, QueryNoTemplate
+from .range import Range, T, dump_range_binary, dump_range_text, fail_dump
+from .range import load_range_binary, load_range_text
+from .._oids import INVALID_OID, TEXT_OID
+from ..adapt import PyFormat, RecursiveDumper, RecursiveLoader
 from .._struct import pack_len, unpack_len
-from ..postgres import INVALID_OID, TEXT_OID
-from .._typeinfo import MultirangeInfo as MultirangeInfo  # exported here
+from .._typeinfo import TypeInfo, TypesRegistry
 
-from .range import Range, T, load_range_text, load_range_binary
-from .range import dump_range_text, dump_range_binary, fail_dump
+if TYPE_CHECKING:
+    from .._connection_base import BaseConnection
+
+
+class MultirangeInfo(TypeInfo):
+    """Manage information about a multirange type."""
+
+    def __init__(
+        self,
+        name: str,
+        oid: int,
+        array_oid: int,
+        *,
+        regtype: str = "",
+        range_oid: int,
+        subtype_oid: int,
+    ):
+        super().__init__(name, oid, array_oid, regtype=regtype)
+        self.range_oid = range_oid
+        self.subtype_oid = subtype_oid
+
+    @classmethod
+    def _get_info_query(cls, conn: BaseConnection[Any]) -> QueryNoTemplate:
+        if conn.info.server_version < 140000:
+            raise e.NotSupportedError(
+                "multirange types are only available from PostgreSQL 14"
+            )
+        return sql.SQL(
+            """\
+SELECT t.typname AS name, t.oid AS oid, t.typarray AS array_oid,
+    t.oid::regtype::text AS regtype,
+    r.rngtypid AS range_oid, r.rngsubtype AS subtype_oid
+FROM pg_type t
+JOIN pg_range r ON t.oid = r.rngmultitypid
+WHERE t.oid = {regtype}
+"""
+        ).format(regtype=cls._to_regtype(conn))
+
+    def _added(self, registry: TypesRegistry) -> None:
+        # Map multiranges ranges and subtypes to info
+        registry._registry[MultirangeInfo, self.range_oid] = self
+        registry._registry[MultirangeInfo, self.subtype_oid] = self
 
 
 class Multirange(MutableSequence[Range[T]]):
@@ -29,7 +75,7 @@ class Multirange(MutableSequence[Range[T]]):
     """
 
     def __init__(self, items: Iterable[Range[T]] = ()):
-        self._ranges: List[Range[T]] = list(map(self._check_type, items))
+        self._ranges: list[Range[T]] = list(map(self._check_type, items))
 
     def _check_type(self, item: Any) -> Range[Any]:
         if not isinstance(item, Range):
@@ -45,14 +91,12 @@ class Multirange(MutableSequence[Range[T]]):
         return f"{{{', '.join(map(str, self._ranges))}}}"
 
     @overload
-    def __getitem__(self, index: int) -> Range[T]:
-        ...
+    def __getitem__(self, index: int) -> Range[T]: ...
 
     @overload
-    def __getitem__(self, index: slice) -> "Multirange[T]":
-        ...
+    def __getitem__(self, index: slice) -> Multirange[T]: ...
 
-    def __getitem__(self, index: Union[int, slice]) -> "Union[Range[T],Multirange[T]]":
+    def __getitem__(self, index: int | slice) -> Range[T] | Multirange[T]:
         if isinstance(index, int):
             return self._ranges[index]
         else:
@@ -62,17 +106,13 @@ class Multirange(MutableSequence[Range[T]]):
         return len(self._ranges)
 
     @overload
-    def __setitem__(self, index: int, value: Range[T]) -> None:
-        ...
+    def __setitem__(self, index: int, value: Range[T]) -> None: ...
 
     @overload
-    def __setitem__(self, index: slice, value: Iterable[Range[T]]) -> None:
-        ...
+    def __setitem__(self, index: slice, value: Iterable[Range[T]]) -> None: ...
 
     def __setitem__(
-        self,
-        index: Union[int, slice],
-        value: Union[Range[T], Iterable[Range[T]]],
+        self, index: int | slice, value: Range[T] | Iterable[Range[T]]
     ) -> None:
         if isinstance(index, int):
             self._check_type(value)
@@ -83,7 +123,7 @@ class Multirange(MutableSequence[Range[T]]):
             value = map(self._check_type, value)
             self._ranges[index] = value
 
-    def __delitem__(self, index: Union[int, slice]) -> None:
+    def __delitem__(self, index: int | slice) -> None:
         del self._ranges[index]
 
     def insert(self, index: int, value: Range[T]) -> None:
@@ -141,9 +181,9 @@ class TimestamptzMultirange(Multirange[datetime]):
 
 
 class BaseMultirangeDumper(RecursiveDumper):
-    def __init__(self, cls: type, context: Optional[AdaptContext] = None):
+    def __init__(self, cls: type, context: AdaptContext | None = None):
         super().__init__(cls, context)
-        self.sub_dumper: Optional[Dumper] = None
+        self.sub_dumper: Dumper | None = None
         self._adapt_format = PyFormat.from_pq(self.format)
 
     def get_key(self, obj: Multirange[Any], format: PyFormat) -> DumperKey:
@@ -151,21 +191,19 @@ class BaseMultirangeDumper(RecursiveDumper):
         if self.cls is not Multirange:
             return self.cls
 
-        item = self._get_item(obj)
-        if item is not None:
+        if (item := self._get_item(obj)) is not None:
             sd = self._tx.get_dumper(item, self._adapt_format)
-            return (self.cls, sd.get_key(item, format))  # type: ignore
+            return (self.cls, sd.get_key(item, format))
         else:
             return (self.cls,)
 
-    def upgrade(self, obj: Multirange[Any], format: PyFormat) -> "BaseMultirangeDumper":
+    def upgrade(self, obj: Multirange[Any], format: PyFormat) -> BaseMultirangeDumper:
         # If we are a subclass whose oid is specified we don't need upgrade
         if self.cls is not Multirange:
             return self
 
-        item = self._get_item(obj)
-        if item is None:
-            return MultirangeDumper(self.cls)
+        if (item := self._get_item(obj)) is None:
+            return self
 
         dumper: BaseMultirangeDumper
         if type(item) is int:
@@ -214,17 +252,16 @@ class MultirangeDumper(BaseMultirangeDumper):
     The dumper can upgrade to one specific for a different range type.
     """
 
-    def dump(self, obj: Multirange[Any]) -> Buffer:
+    def dump(self, obj: Multirange[Any]) -> Buffer | None:
         if not obj:
             return b"{}"
 
-        item = self._get_item(obj)
-        if item is not None:
+        if (item := self._get_item(obj)) is not None:
             dump = self._tx.get_dumper(item, self._adapt_format).dump
         else:
             dump = fail_dump
 
-        out = [b"{"]
+        out: list[Buffer] = [b"{"]
         for r in obj:
             out.append(dump_range_text(r, dump))
             out.append(b",")
@@ -233,17 +270,15 @@ class MultirangeDumper(BaseMultirangeDumper):
 
 
 class MultirangeBinaryDumper(BaseMultirangeDumper):
-
     format = Format.BINARY
 
-    def dump(self, obj: Multirange[Any]) -> Buffer:
-        item = self._get_item(obj)
-        if item is not None:
+    def dump(self, obj: Multirange[Any]) -> Buffer | None:
+        if (item := self._get_item(obj)) is not None:
             dump = self._tx.get_dumper(item, self._adapt_format).dump
         else:
             dump = fail_dump
 
-        out = [pack_len(len(obj))]
+        out: list[Buffer] = [pack_len(len(obj))]
         for r in obj:
             data = dump_range_binary(r, dump)
             out.append(pack_len(len(data)))
@@ -252,10 +287,9 @@ class MultirangeBinaryDumper(BaseMultirangeDumper):
 
 
 class BaseMultirangeLoader(RecursiveLoader, Generic[T]):
-
     subtype_oid: int
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
+    def __init__(self, oid: int, context: AdaptContext | None = None):
         super().__init__(oid, context)
         self._load = self._tx.get_loader(self.subtype_oid, format=self.format).load
 
@@ -307,7 +341,6 @@ _END_INT = ord("}")
 
 
 class MultirangeBinaryLoader(BaseMultirangeLoader[T]):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> Multirange[T]:
@@ -327,7 +360,7 @@ class MultirangeBinaryLoader(BaseMultirangeLoader[T]):
 
 
 def register_multirange(
-    info: MultirangeInfo, context: Optional[AdaptContext] = None
+    info: MultirangeInfo, context: AdaptContext | None = None
 ) -> None:
     """Register the adapters to load and dump a multirange type.
 
@@ -356,20 +389,29 @@ def register_multirange(
     adapters = context.adapters if context else postgres.adapters
 
     # generate and register a customized text loader
-    loader: Type[MultirangeLoader[Any]] = type(
-        f"{info.name.title()}Loader",
-        (MultirangeLoader,),
-        {"subtype_oid": info.subtype_oid},
-    )
+    loader: type[BaseMultirangeLoader[Any]]
+    loader = _make_loader(info.name, info.subtype_oid)
     adapters.register_loader(info.oid, loader)
 
     # generate and register a customized binary loader
-    bloader: Type[MultirangeBinaryLoader[Any]] = type(
-        f"{info.name.title()}BinaryLoader",
-        (MultirangeBinaryLoader,),
-        {"subtype_oid": info.subtype_oid},
+    loader = _make_binary_loader(info.name, info.subtype_oid)
+    adapters.register_loader(info.oid, loader)
+
+
+# Cache all dynamically-generated types to avoid leaks in case the types
+# cannot be GC'd.
+
+
+@cache
+def _make_loader(name: str, oid: int) -> type[MultirangeLoader[Any]]:
+    return type(f"{name.title()}Loader", (MultirangeLoader,), {"subtype_oid": oid})
+
+
+@cache
+def _make_binary_loader(name: str, oid: int) -> type[MultirangeBinaryLoader[Any]]:
+    return type(
+        f"{name.title()}BinaryLoader", (MultirangeBinaryLoader,), {"subtype_oid": oid}
     )
-    adapters.register_loader(info.oid, bloader)
 
 
 # Text dumpers for builtin multirange types wrappers
@@ -378,27 +420,27 @@ def register_multirange(
 
 
 class Int4MultirangeDumper(MultirangeDumper):
-    oid = postgres.types["int4multirange"].oid
+    oid = _oids.INT4MULTIRANGE_OID
 
 
 class Int8MultirangeDumper(MultirangeDumper):
-    oid = postgres.types["int8multirange"].oid
+    oid = _oids.INT8MULTIRANGE_OID
 
 
 class NumericMultirangeDumper(MultirangeDumper):
-    oid = postgres.types["nummultirange"].oid
+    oid = _oids.NUMMULTIRANGE_OID
 
 
 class DateMultirangeDumper(MultirangeDumper):
-    oid = postgres.types["datemultirange"].oid
+    oid = _oids.DATEMULTIRANGE_OID
 
 
 class TimestampMultirangeDumper(MultirangeDumper):
-    oid = postgres.types["tsmultirange"].oid
+    oid = _oids.TSMULTIRANGE_OID
 
 
 class TimestamptzMultirangeDumper(MultirangeDumper):
-    oid = postgres.types["tstzmultirange"].oid
+    oid = _oids.TSTZMULTIRANGE_OID
 
 
 # Binary dumpers for builtin multirange types wrappers
@@ -407,81 +449,81 @@ class TimestamptzMultirangeDumper(MultirangeDumper):
 
 
 class Int4MultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["int4multirange"].oid
+    oid = _oids.INT4MULTIRANGE_OID
 
 
 class Int8MultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["int8multirange"].oid
+    oid = _oids.INT8MULTIRANGE_OID
 
 
 class NumericMultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["nummultirange"].oid
+    oid = _oids.NUMMULTIRANGE_OID
 
 
 class DateMultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["datemultirange"].oid
+    oid = _oids.DATEMULTIRANGE_OID
 
 
 class TimestampMultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["tsmultirange"].oid
+    oid = _oids.TSMULTIRANGE_OID
 
 
 class TimestamptzMultirangeBinaryDumper(MultirangeBinaryDumper):
-    oid = postgres.types["tstzmultirange"].oid
+    oid = _oids.TSTZMULTIRANGE_OID
 
 
 # Text loaders for builtin multirange types
 
 
 class Int4MultirangeLoader(MultirangeLoader[int]):
-    subtype_oid = postgres.types["int4"].oid
+    subtype_oid = _oids.INT4_OID
 
 
 class Int8MultirangeLoader(MultirangeLoader[int]):
-    subtype_oid = postgres.types["int8"].oid
+    subtype_oid = _oids.INT8_OID
 
 
 class NumericMultirangeLoader(MultirangeLoader[Decimal]):
-    subtype_oid = postgres.types["numeric"].oid
+    subtype_oid = _oids.NUMERIC_OID
 
 
 class DateMultirangeLoader(MultirangeLoader[date]):
-    subtype_oid = postgres.types["date"].oid
+    subtype_oid = _oids.DATE_OID
 
 
 class TimestampMultirangeLoader(MultirangeLoader[datetime]):
-    subtype_oid = postgres.types["timestamp"].oid
+    subtype_oid = _oids.TIMESTAMP_OID
 
 
 class TimestampTZMultirangeLoader(MultirangeLoader[datetime]):
-    subtype_oid = postgres.types["timestamptz"].oid
+    subtype_oid = _oids.TIMESTAMPTZ_OID
 
 
 # Binary loaders for builtin multirange types
 
 
 class Int4MultirangeBinaryLoader(MultirangeBinaryLoader[int]):
-    subtype_oid = postgres.types["int4"].oid
+    subtype_oid = _oids.INT4_OID
 
 
 class Int8MultirangeBinaryLoader(MultirangeBinaryLoader[int]):
-    subtype_oid = postgres.types["int8"].oid
+    subtype_oid = _oids.INT8_OID
 
 
 class NumericMultirangeBinaryLoader(MultirangeBinaryLoader[Decimal]):
-    subtype_oid = postgres.types["numeric"].oid
+    subtype_oid = _oids.NUMERIC_OID
 
 
 class DateMultirangeBinaryLoader(MultirangeBinaryLoader[date]):
-    subtype_oid = postgres.types["date"].oid
+    subtype_oid = _oids.DATE_OID
 
 
 class TimestampMultirangeBinaryLoader(MultirangeBinaryLoader[datetime]):
-    subtype_oid = postgres.types["timestamp"].oid
+    subtype_oid = _oids.TIMESTAMP_OID
 
 
 class TimestampTZMultirangeBinaryLoader(MultirangeBinaryLoader[datetime]):
-    subtype_oid = postgres.types["timestamptz"].oid
+    subtype_oid = _oids.TIMESTAMPTZ_OID
 
 
 def register_default_adapters(context: AdaptContext) -> None:
